"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.AbstractRga = void 0;
const clock_1 = require("../../../json-crdt-patch/clock");
const Tree_1 = require("../../../util/trees/Tree");
const isUint8Array_1 = require("../../../util/buffers/isUint8Array");
const util_1 = require("../../../util/trees/util");
const idsCompare = (ts1, ts2) => ts1.sid - ts2.sid || ts1.time - ts2.time;
class AbstractRga {
    constructor(id) {
        this.id = id;
        this.root = undefined;
        this.ids = new Tree_1.Tree(idsCompare);
    }
    ins(ref, id, content) {
        const chunk = this.createChunk(id, content);
        const rootId = this.id;
        const isRootInsert = rootId.time === ref.time && rootId.sid === ref.sid;
        if (isRootInsert) {
            const first = this.first();
            if (!first)
                this.setRoot(chunk);
            else if ((0, clock_1.compare)(first.id, id) < 0)
                this.insertBefore(chunk, first);
            else {
                if ((0, clock_1.containsId)(first.id, first.span, id))
                    return;
                this.insertAfterRef(chunk, ref, first);
            }
        }
        else {
            const at = this.findById(ref);
            if (!at)
                return;
            const offsetInInsertAtChunk = ref.time - at.id.time;
            const atSpan = at.span;
            const needsSplit = offsetInInsertAtChunk + 1 < atSpan;
            if (needsSplit && (0, clock_1.containsId)(at.id, atSpan, id))
                return;
            if (needsSplit && (0, clock_1.compare)(id, (0, clock_1.tick)(ref, 1)) > 0) {
                this.insertInside(chunk, at, offsetInInsertAtChunk + 1);
            }
            else
                this.insertAfterRef(chunk, ref, at);
            this.splay(chunk, 15);
        }
    }
    insAt(position, id, content) {
        const after = position ? this.find(position - 1) : this.id;
        if (!after)
            return undefined;
        this.ins(after, id, content);
        return after;
    }
    delete(spans) {
        const length = spans.length;
        for (let i = 0; i < length; i++)
            this.deleteSpan(spans[i]);
        this.onViewChange();
    }
    deleteSpan(span) {
        const len = span.span;
        const t1 = span.time;
        const t2 = t1 + len - 1;
        const start = this.findById(span);
        if (!start)
            return;
        let chunk = start;
        let last = chunk;
        while (chunk) {
            last = chunk;
            const id = chunk.id;
            const chunkSpan = chunk.span;
            const c1 = id.time;
            const c2 = c1 + chunkSpan - 1;
            if (chunk.del) {
                if (c2 >= t2)
                    break;
                chunk = chunk.s;
                continue;
            }
            const deleteStartsFromLeft = t1 <= c1;
            const deleteStartsInTheMiddle = t1 <= c2;
            if (deleteStartsFromLeft) {
                const deleteFullyContainsChunk = t2 >= c2;
                if (deleteFullyContainsChunk) {
                    chunk.delete();
                    this.dLen(chunk, -chunk.span);
                    if (t2 <= c2)
                        break;
                }
                else {
                    const range = t2 - c1 + 1;
                    const newChunk = this.split(chunk, range);
                    chunk.delete();
                    this.updateLenOne(newChunk);
                    this.dLen(chunk, -chunk.span);
                    break;
                }
            }
            else if (deleteStartsInTheMiddle) {
                const deleteContainsRightSide = t2 >= c2;
                if (deleteContainsRightSide) {
                    const offset = t1 - c1;
                    const newChunk = this.split(chunk, offset);
                    newChunk.delete();
                    newChunk.len = newChunk.r ? newChunk.r.len : 0;
                    this.dLen(chunk, -newChunk.span);
                    if (t2 <= c2)
                        break;
                }
                else {
                    const right = this.split(chunk, t2 - c1 + 1);
                    const mid = this.split(chunk, t1 - c1);
                    mid.delete();
                    this.updateLenOne(right);
                    this.updateLenOne(mid);
                    this.dLen(chunk, -mid.span);
                    break;
                }
            }
            chunk = chunk.s;
        }
        if (last)
            this.mergeTombstones2(start, last);
    }
    find(position) {
        let curr = this.root;
        while (curr) {
            const l = curr.l;
            const leftLength = l ? l.len : 0;
            let span;
            if (position < leftLength)
                curr = l;
            else if (curr.del) {
                position -= leftLength;
                curr = curr.r;
            }
            else if (position < leftLength + (span = curr.span)) {
                const ticks = position - leftLength;
                const id = curr.id;
                return !ticks ? id : new clock_1.Timestamp(id.sid, id.time + ticks);
            }
            else {
                position -= leftLength + span;
                curr = curr.r;
            }
        }
    }
    findChunk(position) {
        let curr = this.root;
        while (curr) {
            const l = curr.l;
            const leftLength = l ? l.len : 0;
            let span;
            if (position < leftLength)
                curr = l;
            else if (curr.del) {
                position -= leftLength;
                curr = curr.r;
            }
            else if (position < leftLength + (span = curr.span)) {
                return [curr, position - leftLength];
            }
            else {
                position -= leftLength + span;
                curr = curr.r;
            }
        }
    }
    findInterval(position, length) {
        const ranges = [];
        if (!length)
            return ranges;
        let curr = this.root;
        let offset = 0;
        while (curr) {
            const leftLength = curr.l ? curr.l.len : 0;
            if (position < leftLength)
                curr = curr.l;
            else if (curr.del) {
                position -= leftLength;
                curr = curr.r;
            }
            else if (position < leftLength + curr.span) {
                offset = position - leftLength;
                break;
            }
            else {
                position -= leftLength + curr.span;
                curr = curr.r;
            }
        }
        if (!curr)
            return ranges;
        if (curr.span >= length + offset) {
            const id = curr.id;
            ranges.push((0, clock_1.tss)(id.sid, id.time + offset, length));
            return ranges;
        }
        const len = curr.span - offset;
        const id = curr.id;
        ranges.push((0, clock_1.tss)(id.sid, id.time + offset, len));
        length -= len;
        curr = this.next(curr);
        if (!curr)
            return ranges;
        do {
            if (curr.del)
                continue;
            const id = curr.id;
            const span = curr.span;
            if (length <= span) {
                ranges.push((0, clock_1.tss)(id.sid, id.time, length));
                return ranges;
            }
            ranges.push((0, clock_1.tss)(id.sid, id.time, span));
            length -= span;
        } while ((curr = this.next(curr)) && length > 0);
        return ranges;
    }
    first() {
        let curr = this.root;
        while (curr)
            if (curr.l)
                curr = curr.l;
            else
                return curr;
        return curr;
    }
    next(curr) {
        if (curr.r) {
            curr = curr.r;
            while (curr.l)
                curr = curr.l;
            return curr;
        }
        let p = curr.p;
        while (p && p.r === curr) {
            curr = p;
            p = p.p;
        }
        return p;
    }
    prev(curr) {
        if (curr.l) {
            curr = curr.l;
            while (curr.r)
                curr = curr.r;
            return curr;
        }
        let p = curr.p;
        while (p && p.l === curr) {
            curr = p;
            p = p.p;
        }
        return p;
    }
    length() {
        const root = this.root;
        return root ? root.len : 0;
    }
    size() {
        return this.ids.size;
    }
    pos(chunk) {
        const p = chunk.p;
        if (!p)
            return chunk.l ? chunk.l.len : 0;
        return (this.pos(p) +
            (p.r === chunk ? (p.del ? 0 : p.span) : (chunk.del ? 0 : -chunk.span) + (chunk.r ? -chunk.r.len : 0)));
    }
    setRoot(chunk) {
        this.root = chunk;
        this.insertId(chunk);
        this.onViewChange();
    }
    insertBefore(chunk, before) {
        const l = before.l;
        before.l = chunk;
        chunk.l = l;
        chunk.p = before;
        if (l)
            l.p = chunk;
        this.updateLenOne(chunk);
        this.dLen(before, chunk.span);
        this.insertId(chunk);
        this.onViewChange();
    }
    insertAfter(chunk, after) {
        const r = after.r;
        after.r = chunk;
        chunk.r = r;
        chunk.p = after;
        if (r)
            r.p = chunk;
        this.updateLenOne(chunk);
        this.dLen(after, chunk.span);
        this.insertId(chunk);
        this.onViewChange();
    }
    insertAfterRef(chunk, ref, left) {
        const id = chunk.id;
        const sid = id.sid;
        const time = id.time;
        let isSplit = false;
        for (;;) {
            const leftId = left.id;
            const leftNextTick = leftId.time + left.span;
            if (!left.s) {
                isSplit = leftId.sid === sid && leftNextTick === time && leftNextTick - 1 === ref.time;
                if (isSplit)
                    left.s = chunk;
            }
            const right = this.next(left);
            if (!right)
                break;
            const cmp = (0, clock_1.compare)(right.id, id);
            if (cmp < 0)
                break;
            if (!cmp)
                return;
            left = right;
        }
        if (isSplit && !left.del) {
            this.mergeContent(left, chunk.data);
            left.s = undefined;
        }
        else
            this.insertAfter(chunk, left);
    }
    mergeContent(chunk, content) {
        const span1 = chunk.span;
        chunk.merge(content);
        this.dLen(chunk, chunk.span - span1);
        this.onViewChange();
        return;
    }
    insertInside(chunk, at, offset) {
        const p = at.p;
        const l = at.l;
        const r = at.r;
        const s = at.s;
        const len = at.len;
        const at2 = at.split(offset);
        at.s = at2;
        at2.s = s;
        at.l = at.r = at2.l = at2.r = undefined;
        at2.l = undefined;
        chunk.p = p;
        if (!l) {
            chunk.l = at;
            at.p = chunk;
        }
        else {
            chunk.l = l;
            l.p = chunk;
            const a = l.r;
            l.r = at;
            at.p = l;
            at.l = a;
            if (a)
                a.p = at;
        }
        if (!r) {
            chunk.r = at2;
            at2.p = chunk;
        }
        else {
            chunk.r = r;
            r.p = chunk;
            const b = r.l;
            r.l = at2;
            at2.p = r;
            at2.r = b;
            if (b)
                b.p = at2;
        }
        if (!p)
            this.root = chunk;
        else if (p.l === at)
            p.l = chunk;
        else
            p.r = chunk;
        this.updateLenOne(at);
        this.updateLenOne(at2);
        if (l)
            l.len = (l.l ? l.l.len : 0) + at.len + (l.del ? 0 : l.span);
        if (r)
            r.len = (r.r ? r.r.len : 0) + at2.len + (r.del ? 0 : r.span);
        chunk.len = len + chunk.span;
        const span = chunk.span;
        let curr = chunk.p;
        while (curr) {
            curr.len += span;
            curr = curr.p;
        }
        this.insertId(at2);
        this.insertIdFast(chunk);
        this.onViewChange();
    }
    split(chunk, ticks) {
        const s = chunk.s;
        const newChunk = chunk.split(ticks);
        const r = chunk.r;
        chunk.s = newChunk;
        newChunk.r = r;
        newChunk.s = s;
        chunk.r = newChunk;
        newChunk.p = chunk;
        this.insertId(newChunk);
        if (r)
            r.p = newChunk;
        return newChunk;
    }
    mergeTombstones(ch1, ch2) {
        if (!ch1.del || !ch2.del)
            return false;
        const id1 = ch1.id;
        const id2 = ch2.id;
        if (id1.sid !== id2.sid)
            return false;
        if (id1.time + ch1.span !== id2.time)
            return false;
        ch1.s = ch2.s;
        ch1.span += ch2.span;
        this.deleteChunk(ch2);
        return true;
    }
    mergeTombstones2(start, end) {
        let curr = start;
        while (curr) {
            const next = this.next(curr);
            if (!next)
                break;
            const merged = this.mergeTombstones(curr, next);
            if (!merged) {
                if (next === end) {
                    if (next) {
                        const n = this.next(next);
                        if (n)
                            this.mergeTombstones(next, n);
                    }
                    break;
                }
                curr = curr.s;
            }
        }
        const prev = this.prev(start);
        if (prev)
            this.mergeTombstones(prev, start);
    }
    deleteChunk(chunk) {
        this.deleteId(chunk);
        const p = chunk.p;
        const l = chunk.l;
        const r = chunk.r;
        if (!l && !r) {
            if (!p)
                this.root = undefined;
            else {
                if (p.l === chunk)
                    p.l = undefined;
                else
                    p.r = undefined;
            }
        }
        else if (l && r) {
            let mostRightChildFromLeft = l;
            while (mostRightChildFromLeft.r)
                mostRightChildFromLeft = mostRightChildFromLeft.r;
            mostRightChildFromLeft.r = r;
            r.p = mostRightChildFromLeft;
            const rLen = r.len;
            let curr;
            curr = mostRightChildFromLeft;
            if (!p) {
                this.root = l;
                l.p = undefined;
            }
            else {
                if (p.l === chunk)
                    p.l = l;
                else
                    p.r = l;
                l.p = p;
            }
            while (curr && curr !== p) {
                curr.len += rLen;
                curr = curr.p;
            }
        }
        else {
            const child = (l || r);
            child.p = p;
            if (!p)
                this.root = child;
            else if (p.l === chunk)
                p.l = child;
            else
                p.r = child;
        }
    }
    insertId(chunk) {
        this.ids.set(chunk.id, chunk);
    }
    insertIdFast(chunk) {
        this.ids.setFast(chunk.id, chunk);
    }
    deleteId(chunk) {
        this.ids.delete(chunk.id);
    }
    findById(id) {
        const chunk = this.ids.getOrNextLower(id);
        if (!chunk)
            return undefined;
        const chunkId = chunk.id;
        if (chunkId.sid !== id.sid)
            return undefined;
        const offset = id.time - chunkId.time;
        return offset < chunk.span ? chunk : undefined;
    }
    dLen(chunk, delta) {
        while (chunk) {
            chunk.len += delta;
            chunk = chunk.p;
        }
    }
    updateLenOne(chunk) {
        const l = chunk.l;
        const r = chunk.r;
        chunk.len = (chunk.del ? 0 : chunk.span) + (l ? l.len : 0) + (r ? r.len : 0);
    }
    splay(chunk, repeat) {
        const p = chunk.p;
        if (!p)
            return;
        const pp = p.p;
        const l2 = p.l === chunk;
        if (!pp) {
            if (l2)
                this.root = (0, util_1.rSplay)(chunk, p);
            else
                this.root = (0, util_1.lSplay)(chunk, p);
            this.updateLenOne(p);
            this.updateLenOne(chunk);
            return;
        }
        const l1 = pp.l === p;
        if (l1 && l2)
            this.root = (0, util_1.llSplay)(this.root, chunk, p, pp);
        else if (!l1 && !l2)
            this.root = (0, util_1.rrSplay)(this.root, chunk, p, pp);
        else if (l1 && !l2)
            this.root = (0, util_1.lrSplay)(this.root, chunk, p, pp);
        else
            this.root = (0, util_1.rlSplay)(this.root, chunk, p, pp);
        this.updateLenOne(pp);
        this.updateLenOne(p);
        this.updateLenOne(chunk);
        if (repeat > 1)
            this.splay(chunk, repeat - 1);
    }
    toString(tab = '') {
        const view = this.view();
        let value = '';
        if ((0, isUint8Array_1.isUint8Array)(view))
            value += ` { ${view.toString()} }`;
        else if (typeof view === 'string')
            value += `{ ${view.length > 32 ? JSON.stringify(view.substring(0, 32)) + '...' : JSON.stringify(view)} }`;
        return `${this.constructor.name} ${(0, clock_1.toDisplayString)(this.id)} ${value}${this.root ? this.toStringChunk(this.root, tab, '') : '[empty]'}`;
    }
    toStringChunk(chunk, tab, side) {
        let str = `\n${tab}${this.formatChunk(chunk, tab, side)}`;
        if (chunk.l)
            str += this.toStringChunk(chunk.l, tab + '  ', 'l');
        if (chunk.r)
            str += this.toStringChunk(chunk.r, tab + '  ', 'r');
        return str;
    }
    formatChunk(chunk, tab, side) {
        let str = `${side === 'l' ? '<-' : side === 'r' ? '->' : '└─'} ${chunk.constructor.name} ${(0, clock_1.toDisplayString)(chunk.id)}!${chunk.span} len:${chunk.len}`;
        if (chunk.del)
            str += ` [${chunk.span}]`;
        else {
            if ((0, isUint8Array_1.isUint8Array)(chunk.data))
                str += ` { ${chunk.data.toString()} }`;
            else if (typeof chunk.data === 'string')
                str += ` { ${chunk.data.length > 32 ? JSON.stringify(chunk.data.substring(0, 32)) + '...' : JSON.stringify(chunk.data)} }`;
        }
        if (chunk.s)
            str += ` ${chunk.s ? `---> ${(0, clock_1.toDisplayString)(chunk.s.id)}` : ''}`;
        return str;
    }
    iterator() {
        let curr = this.first();
        return () => {
            const res = curr;
            if (curr)
                curr = this.next(curr);
            return res;
        };
    }
    ingest(size, next) {
        if (size < 1)
            return;
        const splitLeftChunks = new Map();
        this.root = this._ingest(size, () => {
            const chunk = next();
            const id = chunk.id;
            const key = id.sid + '.' + id.time;
            const split = splitLeftChunks.get(key);
            if (split) {
                split.s = chunk;
                splitLeftChunks.delete(key);
            }
            const nextStampAfterSpan = (0, clock_1.tick)(id, chunk.span);
            splitLeftChunks.set(nextStampAfterSpan.sid + '.' + nextStampAfterSpan.time, chunk);
            return chunk;
        });
    }
    _ingest(size, next) {
        const leftSize = size >> 1;
        const rightSize = size - leftSize - 1;
        const c1 = leftSize > 0 ? this._ingest(leftSize, next) : undefined;
        const c2 = next();
        if (c1) {
            c2.l = c1;
            c1.p = c2;
        }
        const c3 = rightSize > 0 ? this._ingest(rightSize, next) : undefined;
        if (c3) {
            c2.r = c3;
            c3.p = c2;
        }
        this.updateLenOne(c2);
        this.insertId(c2);
        return c2;
    }
}
exports.AbstractRga = AbstractRga;
