"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.CborEncoder = void 0;
const isFloat32_1 = require("../../util/buffers/isFloat32");
const isUint8Array_1 = require("../../util/buffers/isUint8Array");
const CborEncoderFast_1 = require("./CborEncoderFast");
const isSafeInteger = Number.isSafeInteger;
class CborEncoder extends CborEncoderFast_1.CborEncoderFast {
    encode(value) {
        this.writer.reset();
        this.writeAny(value);
        return this.writer.flush();
    }
    writeAny(value) {
        switch (value) {
            case null:
                return this.writer.u8(0xf6);
            case false:
                return this.writer.u8(0xf4);
            case true:
                return this.writer.u8(0xf5);
            case undefined:
                return this.writer.u8(0xf7);
        }
        if (value instanceof Array)
            return this.writeArr(value);
        switch (typeof value) {
            case 'number':
                return this.encodeNumber(value);
            case 'string':
                return this.writeStr(value);
            case 'object': {
                if ((0, isUint8Array_1.isUint8Array)(value))
                    return this.writeBin(value);
                if (value instanceof Map)
                    return this.writeMap(value);
                return this.writeObj(value);
            }
            case 'bigint':
                return this.encodeBigInteger(value);
        }
    }
    encodeNumber(num) {
        if (isSafeInteger(num))
            this.encodeInteger(num);
        else if (typeof num === 'bigint')
            this.encodeBigInteger(num);
        else
            this.writeFloat(num);
    }
    encodeBigInteger(int) {
        if (int >= 0)
            this.encodeBigUint(int);
        else
            this.encodeBigNint(int);
    }
    encodeBigUint(uint) {
        switch (true) {
            case uint <= 23:
                this.writer.u8(0 + Number(uint));
                break;
            case uint <= 0xff:
                this.writer.u16((0x18 << 8) + Number(uint));
                break;
            case uint <= 0xffff:
                this.writer.u8u16(0x19, Number(uint));
                break;
            case uint <= 0xffffffff:
                this.writer.u8u32(0x1a, Number(uint));
                break;
            default: {
                this.writer.u8u64(0x1b, uint);
                break;
            }
        }
    }
    encodeBigNint(int) {
        const uint = -BigInt(1) - int;
        switch (true) {
            case uint <= 24:
                this.writer.u8(32 + Number(uint));
                break;
            case uint <= 0xff:
                this.writer.u16((0x38 << 8) + Number(uint));
                break;
            case uint <= 0xffff:
                this.writer.u8u16(0x39, Number(uint));
                break;
            case uint <= 0xffffffff:
                this.writer.u8u32(0x3a, Number(uint));
                break;
            default:
                this.writer.u8u64(0x3b, uint);
                break;
        }
    }
    writeFloat(float) {
        if ((0, isFloat32_1.isFloat32)(float))
            this.writer.u8f32(0xfa, float);
        else
            this.writer.u8f64(0xfb, float);
    }
    writeMap(map) {
        this.writeMapHdr(map.size);
        map.forEach((value, key) => {
            this.writeAny(key);
            this.writeAny(value);
        });
    }
}
exports.CborEncoder = CborEncoder;
