"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.WorkerRuntime = void 0;
const tslib_1 = require("tslib");
const WorkerResponse_1 = require("../WorkerResponse");
const WorkerChannel_1 = require("../WorkerChannel");
class WorkerRuntime {
    constructor(port) {
        this.port = port;
        this.fn = [];
        this.channels = new Map();
        this.port.on('message', (msg) => {
            if (msg instanceof Array) {
                if (msg[0] instanceof Array)
                    this.onChannelMsg(msg);
                else
                    this.onRequest(msg);
            }
            else {
                switch (msg.type) {
                    case 'load':
                        this.onLoad(msg);
                }
            }
        });
    }
    onChannelMsg(msg) {
        const [[seq], data] = msg;
        const callback = this.channels.get(seq);
        if (callback)
            callback(data);
    }
    onRequest([seq, method, data]) {
        const fn = this.fn[method];
        if (!fn) {
            return;
        }
        if (fn instanceof WorkerChannel_1.WorkerChannel)
            this.onCh(seq, fn, data);
        else
            this.onFn(seq, fn, data);
    }
    onFn(seq, fn, data) {
        try {
            const result = fn(data);
            if (result instanceof Promise) {
                result
                    .then((result) => {
                    this.sendResponse(seq, result);
                })
                    .catch((error) => {
                    this.sendError(seq, error);
                });
            }
            else {
                this.sendResponse(seq, result);
            }
        }
        catch (error) {
            this.sendError(seq, error);
        }
    }
    onCh(seq, fn, data) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            let closed = false;
            try {
                const send = (data) => {
                    if (closed)
                        throw new Error('CHANNEL_CLOSED');
                    let transferList;
                    if (data instanceof WorkerResponse_1.WorkerResponse) {
                        transferList = data.transferList;
                        data = data.data;
                    }
                    const msg = [[seq], data];
                    this.port.postMessage(msg, transferList);
                };
                const take = (callback) => {
                    if (closed)
                        throw new Error('CHANNEL_CLOSED');
                    this.channels.set(seq, callback);
                };
                const response = yield fn.fn(data, take, send);
                closed = true;
                this.channels.delete(seq);
                this.sendResponse(seq, response);
            }
            catch (error) {
                closed = true;
                this.channels.delete(seq);
                this.sendError(seq, error);
            }
        });
    }
    sendResponse(seq, response) {
        let transferList;
        if (response instanceof WorkerResponse_1.WorkerResponse) {
            transferList = response.transferList;
            response = response.data;
        }
        const msg = [seq, response];
        this.port.postMessage(msg, transferList);
    }
    sendError(seq, error) {
        try {
            const response = [seq, error, true];
            this.port.postMessage(response);
        }
        catch (_a) {
            const response = [seq, 0, true];
            this.port.postMessage(response);
        }
    }
    onLoad({ file }) {
        const { methods } = require(file);
        const keys = Object.keys(methods).sort();
        const list = [];
        for (const key of keys) {
            const isChannel = methods[key] instanceof WorkerChannel_1.WorkerChannel;
            const fn = typeof methods[key] === 'function'
                ? methods[key]
                : isChannel
                    ? methods[key]
                    : () => new WorkerResponse_1.WorkerResponse(methods[key], []);
            const id = this.fn.length;
            list.push([id, key, isChannel ? 'ch' : 'fn']);
            this.fn.push(fn);
        }
        const response = {
            type: 'loaded',
            list,
        };
        this.port.postMessage(response);
    }
    ready() {
        const msg = { type: 'ready' };
        this.port.postMessage(msg);
    }
}
exports.WorkerRuntime = WorkerRuntime;
