import { ArrayRga, ArrayChunk } from '../../../types/rga-array/ArrayRga';
import { BinaryRga, BinaryChunk } from '../../../types/rga-binary/BinaryRga';
import { ClockDecoder } from '../../../../json-crdt-patch/codec/clock/ClockDecoder';
import { Const } from '../../../types/const/Const';
import { CrdtDecoder } from '../../../../json-crdt-patch/util/binary/CrdtDecoder';
import { Timestamp } from '../../../../json-crdt-patch/clock';
import { Model, UNDEFINED } from '../../../model';
import { MsgPackDecoderFast } from '../../../../json-pack/msgpack';
import { ObjectLww } from '../../../types/lww-object/ObjectLww';
import { RootLww } from '../../../types/lww-root/RootLww';
import { StringRga, StringChunk } from '../../../types/rga-string/StringRga';
import { ValueLww } from '../../../types/lww-value/ValueLww';
export class Decoder extends MsgPackDecoderFast {
    doc;
    clockDecoder;
    time = -1;
    constructor() {
        super(new CrdtDecoder());
    }
    decode(data) {
        delete this.clockDecoder;
        this.time = -1;
        this.reader.reset(data);
        const isServerTime = this.reader.u8() === 0;
        if (isServerTime) {
            const time = (this.time = this.reader.vu57());
            this.doc = Model.withServerClock(time);
        }
        else {
            this.decodeClockTable();
            const clock = this.clockDecoder.clock;
            this.doc = Model.withLogicalClock(clock);
        }
        this.doc.root = new RootLww(this.doc, this.cRoot().id);
        delete this.clockDecoder;
        return this.doc;
    }
    decodeClockTable() {
        const clockTableOffset = this.reader.u32();
        const offset = this.reader.x;
        this.reader.x += clockTableOffset;
        const length = this.reader.vu57();
        const [sessionId, time] = this.reader.u53vu39();
        this.clockDecoder = new ClockDecoder(sessionId, time);
        for (let i = 1; i < length; i++) {
            const [sid, time] = this.reader.u53vu39();
            this.clockDecoder.pushTuple(sid, time);
        }
        this.reader.x = offset;
    }
    ts() {
        const decoderTime = this.time;
        const isLogical = decoderTime < 0;
        if (isLogical) {
            const [sessionIndex, timeDiff] = this.reader.id();
            return this.clockDecoder.decodeId(sessionIndex, timeDiff);
        }
        else {
            return new Timestamp(1, this.reader.vu57());
        }
    }
    cRoot() {
        const peek = this.reader.uint8[this.reader.x];
        return !peek ? UNDEFINED : this.cNode();
    }
    cNode() {
        const id = this.ts();
        const byte = this.reader.u8();
        if (byte <= 0b10001111)
            return this.cObj(id, byte & 0b1111);
        else if (byte <= 0b10011111)
            return this.cArr(id, byte & 0b1111);
        else if (byte <= 0b10111111)
            return this.cStr(id, byte & 0b11111);
        else {
            switch (byte) {
                case 0xc4:
                    return this.cBin(id, this.reader.u8());
                case 0xc5:
                    return this.cBin(id, this.reader.u16());
                case 0xc6:
                    return this.cBin(id, this.reader.u32());
                case 0xd4:
                    return this.cConst(id);
                case 0xd6:
                    return this.cVal(id);
                case 0xde:
                    return this.cObj(id, this.reader.u16());
                case 0xdf:
                    return this.cObj(id, this.reader.u32());
                case 0xdc:
                    return this.cArr(id, this.reader.u16());
                case 0xdd:
                    return this.cArr(id, this.reader.u32());
                case 0xd9:
                    return this.cStr(id, this.reader.u8());
                case 0xda:
                    return this.cStr(id, this.reader.u16());
                case 0xdb:
                    return this.cStr(id, this.reader.u32());
            }
        }
        throw new Error('UNKNOWN_NODE');
    }
    cObj(id, length) {
        const obj = new ObjectLww(this.doc, id);
        for (let i = 0; i < length; i++)
            this.cObjChunk(obj);
        this.doc.index.set(obj);
        return obj;
    }
    cObjChunk(obj) {
        const key = this.val();
        obj.put(key, this.cNode().id);
    }
    cArr(id, length) {
        const obj = new ArrayRga(this.doc, id);
        obj.ingest(length, this.cArrChunk);
        this.doc.index.set(obj);
        return obj;
    }
    cArrChunk = () => {
        const [deleted, length] = this.reader.b1vu28();
        const id = this.ts();
        if (deleted)
            return new ArrayChunk(id, length, undefined);
        const ids = [];
        for (let i = 0; i < length; i++)
            ids.push(this.cNode().id);
        return new ArrayChunk(id, length, ids);
    };
    cStr(id, length) {
        const node = new StringRga(id);
        if (length)
            node.ingest(length, this.cStrChunk);
        this.doc.index.set(node);
        return node;
    }
    cStrChunk = () => {
        const id = this.ts();
        const isTombstone = this.reader.uint8[this.reader.x] === 0;
        if (isTombstone) {
            this.reader.x++;
            const length = this.reader.vu39();
            return new StringChunk(id, length, '');
        }
        const text = this.val();
        return new StringChunk(id, text.length, text);
    };
    cBin(id, length) {
        const node = new BinaryRga(id);
        if (length)
            node.ingest(length, this.cBinChunk);
        this.doc.index.set(node);
        return node;
    }
    cBinChunk = () => {
        const [deleted, length] = this.reader.b1vu28();
        const id = this.ts();
        if (deleted)
            return new BinaryChunk(id, length, undefined);
        else
            return new BinaryChunk(id, length, this.bin(length));
    };
    cConst(id) {
        const obj = new Const(id, this.val());
        this.doc.index.set(obj);
        return obj;
    }
    cVal(id) {
        const val = this.cNode();
        const obj = new ValueLww(this.doc, id, val.id);
        this.doc.index.set(obj);
        return obj;
    }
}
