import { Writer } from '../../util/buffers/Writer';
export class MsgPackEncoderFast {
    writer;
    constructor(writer = new Writer()) {
        this.writer = writer;
    }
    encode(json) {
        this.writer.reset();
        this.encodeAny(json);
        return this.writer.flush();
    }
    encodeAny(json) {
        switch (json) {
            case null:
                return this.writer.u8(0xc0);
            case false:
                return this.writer.u8(0xc2);
            case true:
                return this.writer.u8(0xc3);
        }
        if (json instanceof Array)
            return this.encodeArray(json);
        switch (typeof json) {
            case 'number':
                return this.encodeNumber(json);
            case 'string':
                return this.encodeString(json);
            case 'object':
                return this.encodeObject(json);
        }
    }
    encodeFloat64(num) {
        this.writer.u8f64(0xcb, num);
    }
    encodeUnsignedInteger(num) {
        const writer = this.writer;
        this.writer.ensureCapacity(5);
        const uint8 = writer.uint8;
        if (num <= 0b1111111) {
            uint8[writer.x++] = num;
        }
        else if (num <= 0xffff) {
            uint8[writer.x++] = 0xcd;
            writer.view.setUint16(writer.x, num);
            writer.x += 2;
        }
        else {
            uint8[writer.x++] = 0xce;
            writer.view.setUint32(writer.x, num);
            writer.x += 4;
        }
    }
    encodeNegativeInteger(num) {
        const writer = this.writer;
        this.writer.ensureCapacity(5);
        const uint8 = writer.uint8;
        if (num >= -0x20) {
            uint8[writer.x++] = 0x100 + num;
        }
        else if (num >= -0x8000) {
            uint8[writer.x++] = 0xd1;
            writer.view.setInt16(writer.x, num);
            writer.x += 2;
        }
        else {
            uint8[writer.x++] = 0xd2;
            writer.view.setInt32(writer.x, num);
            writer.x += 4;
        }
    }
    encodeNumber(num) {
        if (num >>> 0 === num)
            return this.encodeUnsignedInteger(num);
        if (num >> 0 === num)
            return this.encodeNegativeInteger(num);
        this.encodeFloat64(num);
    }
    encodeNull() {
        this.writer.u8(0xc0);
    }
    encodeTrue() {
        this.writer.u8(0xc3);
    }
    encodeFalse() {
        this.writer.u8(0xc2);
    }
    encodeBoolean(bool) {
        if (bool)
            this.writer.u8(0xc3);
        else
            this.writer.u8(0xc2);
    }
    encodeStringHeader(length) {
        if (length <= 0b11111)
            this.writer.u8(0b10100000 | length);
        else if (length <= 0xff) {
            this.writer.u8(0xd9);
            this.writer.u8(length);
        }
        else if (length <= 0xffff) {
            this.writer.u8(0xda);
            this.writer.u16(length);
        }
        else if (length <= 0xffffffff) {
            this.writer.u8(0xdb);
            this.writer.u32(length);
        }
    }
    encodeString(str) {
        const writer = this.writer;
        const length = str.length;
        const maxSize = length * 4;
        writer.ensureCapacity(5 + maxSize);
        const uint8 = writer.uint8;
        let lengthOffset = writer.x;
        if (maxSize <= 0b11111)
            writer.x++;
        else if (maxSize <= 0xff) {
            uint8[writer.x++] = 0xd9;
            lengthOffset = writer.x;
            writer.x++;
        }
        else if (maxSize <= 0xffff) {
            uint8[writer.x++] = 0xda;
            lengthOffset = writer.x;
            writer.x += 2;
        }
        else {
            uint8[writer.x++] = 0xdb;
            lengthOffset = writer.x;
            writer.x += 4;
        }
        const bytesWritten = this.writer.utf8(str);
        if (maxSize <= 0b11111)
            uint8[lengthOffset] = 0b10100000 | bytesWritten;
        else if (maxSize <= 0xff)
            uint8[lengthOffset] = bytesWritten;
        else if (maxSize <= 0xffff)
            writer.view.setUint16(lengthOffset, bytesWritten);
        else
            writer.view.setUint32(lengthOffset, bytesWritten);
    }
    encodeAsciiString(str) {
        const length = str.length;
        this.encodeStringHeader(length);
        this.writer.ensureCapacity(length);
        const uint8 = this.writer.uint8;
        let offset = this.writer.x;
        let pos = 0;
        while (pos < length)
            uint8[offset++] = str.charCodeAt(pos++);
        this.writer.x = offset;
    }
    encodeArrayHeader(length) {
        if (length <= 0b1111)
            this.writer.u8(0b10010000 | length);
        else if (length <= 0xffff) {
            this.writer.u8u16(0xdc, length);
        }
        else if (length <= 0xffffffff) {
            this.writer.u8u32(0xdd, length);
        }
    }
    encodeArray(arr) {
        const length = arr.length;
        if (length <= 0b1111)
            this.writer.u8(0b10010000 | length);
        else if (length <= 0xffff) {
            this.writer.u8u16(0xdc, length);
        }
        else if (length <= 0xffffffff) {
            this.writer.u8u32(0xdd, length);
        }
        else
            return;
        for (let i = 0; i < length; i++)
            this.encodeAny(arr[i]);
    }
    encodeObjectHeader(length) {
        if (length <= 0b1111)
            this.writer.u8(0b10000000 | length);
        else if (length <= 0xffff) {
            this.writer.u8u16(0xde, length);
        }
        else if (length <= 0xffffffff) {
            this.writer.u8u32(0xdf, length);
        }
    }
    encodeObject(obj) {
        const keys = Object.keys(obj);
        const length = keys.length;
        this.encodeObjectHeader(length);
        for (let i = 0; i < length; i++) {
            const key = keys[i];
            this.encodeString(key);
            this.encodeAny(obj[key]);
        }
    }
    encodeExtHeader(type, length) {
        switch (length) {
            case 1:
                this.writer.u16((0xd4 << 8) | type);
                break;
            case 2:
                this.writer.u16((0xd5 << 8) | type);
                break;
            case 4:
                this.writer.u16((0xd6 << 8) | type);
                break;
            case 8:
                this.writer.u16((0xd7 << 8) | type);
                break;
            case 16:
                this.writer.u16((0xd8 << 8) | type);
                break;
            default:
                if (length <= 0xff) {
                    this.writer.u16((0xc7 << 8) | length);
                    this.writer.u8(type);
                }
                else if (length <= 0xffff) {
                    this.writer.u8u16(0xc8, length);
                    this.writer.u8(type);
                }
                else if (length <= 0xffffffff) {
                    this.writer.u8u32(0xc9, length);
                    this.writer.u8(type);
                }
        }
    }
    encodeExt(ext) {
        const { tag: type, val: buf } = ext;
        const length = buf.byteLength;
        this.encodeExtHeader(type, length);
        this.writer.buf(buf, length);
    }
    encodeBinaryHeader(length) {
        if (length <= 0xff)
            this.writer.u16((0xc4 << 8) | length);
        else if (length <= 0xffff) {
            this.writer.u8u16(0xc5, length);
        }
        else if (length <= 0xffffffff) {
            this.writer.u8u32(0xc6, length);
        }
    }
    encodeBinary(buf) {
        const length = buf.byteLength;
        this.encodeBinaryHeader(length);
        this.writer.buf(buf, length);
    }
}
