"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Decoder = void 0;
var tslib_1 = require("tslib");
var ArrayRga_1 = require("../../../types/rga-array/ArrayRga");
var BinaryRga_1 = require("../../../types/rga-binary/BinaryRga");
var ClockTable_1 = require("../../../../json-crdt-patch/codec/clock/ClockTable");
var Const_1 = require("../../../types/const/Const");
var CrdtDecoder_1 = require("../../../../json-crdt-patch/util/binary/CrdtDecoder");
var clock_1 = require("../../../../json-crdt-patch/clock");
var model_1 = require("../../../model");
var msgpack_1 = require("../../../../json-pack/msgpack");
var ObjectLww_1 = require("../../../types/lww-object/ObjectLww");
var StringRga_1 = require("../../../types/rga-string/StringRga");
var ValueLww_1 = require("../../../types/lww-value/ValueLww");
var Decoder = (function () {
    function Decoder() {
        this.dec = new msgpack_1.MsgPackDecoderFast(new CrdtDecoder_1.CrdtDecoder());
    }
    Decoder.prototype.decode = function (fields, ModelConstructor) {
        if (ModelConstructor === void 0) { ModelConstructor = model_1.Model; }
        var reader = this.dec.reader;
        reader.reset(fields.c);
        var clockTable = (this.clockTable = ClockTable_1.ClockTable.decode(reader));
        return this.decodeFields(clockTable, fields, ModelConstructor);
    };
    Decoder.prototype.decodeFields = function (clockTable, fields, ModelConstructor) {
        if (ModelConstructor === void 0) { ModelConstructor = model_1.Model; }
        var reader = this.dec.reader;
        var firstClock = clockTable.byIdx[0];
        var vectorClock = new clock_1.VectorClock(firstClock.sid, firstClock.time + 1);
        var doc = (this.doc = new ModelConstructor(vectorClock));
        var root = fields.r;
        if (root && root.length) {
            reader.reset(root);
            var rootValue = this.ts();
            doc.root.set(rootValue);
        }
        var docIndex = doc.index;
        for (var field in fields) {
            if (field.length < 3)
                continue;
            var arr = fields[field];
            var id = clockTable.parseField(field);
            reader.reset(arr);
            var node = this.decodeNode(id);
            docIndex.set(node);
        }
        return doc;
    };
    Decoder.prototype.ts = function () {
        var _a = tslib_1.__read(this.dec.reader.id(), 2), sessionIndex = _a[0], timeDiff = _a[1];
        return new clock_1.Timestamp(this.clockTable.byIdx[sessionIndex].sid, timeDiff);
    };
    Decoder.prototype.decodeNode = function (id) {
        var reader = this.dec.reader;
        var byte = reader.u8();
        if (byte <= 143)
            return this.cObj(id, byte & 15);
        else if (byte <= 159)
            return this.cArr(id, byte & 15);
        else if (byte <= 191)
            return this.cStr(id, byte & 31);
        else {
            switch (byte) {
                case 0xc4:
                    return this.cBin(id, reader.u8());
                case 0xc5:
                    return this.cBin(id, reader.u16());
                case 0xc6:
                    return this.cBin(id, reader.u32());
                case 0xd4:
                    return this.cConst(id);
                case 0xd6:
                    return this.cVal(id);
                case 0xde:
                    return this.cObj(id, reader.u16());
                case 0xdf:
                    return this.cObj(id, reader.u32());
                case 0xdc:
                    return this.cArr(id, reader.u16());
                case 0xdd:
                    return this.cArr(id, reader.u32());
                case 0xd9:
                    return this.cStr(id, reader.u8());
                case 0xda:
                    return this.cStr(id, reader.u16());
                case 0xdb:
                    return this.cStr(id, reader.u32());
            }
        }
        return model_1.UNDEFINED;
    };
    Decoder.prototype.cConst = function (id) {
        var val = this.dec.val();
        return new Const_1.Const(id, val);
    };
    Decoder.prototype.cVal = function (id) {
        var val = this.ts();
        return new ValueLww_1.ValueLww(this.doc, id, val);
    };
    Decoder.prototype.cObj = function (id, length) {
        var decoder = this.dec;
        var obj = new ObjectLww_1.ObjectLww(this.doc, id);
        var keys = obj.latest;
        for (var i = 0; i < length; i++) {
            var key = String(decoder.val());
            var val = this.ts();
            keys.set(key, val);
        }
        return obj;
    };
    Decoder.prototype.cStr = function (id, length) {
        var _this = this;
        var decoder = this.dec;
        var node = new StringRga_1.StringRga(id);
        node.ingest(length, function () {
            var chunkId = _this.ts();
            var val = decoder.val();
            if (typeof val === 'number')
                return new StringRga_1.StringChunk(chunkId, val, '');
            var data = String(val);
            return new StringRga_1.StringChunk(chunkId, data.length, data);
        });
        return node;
    };
    Decoder.prototype.cBin = function (id, length) {
        var _this = this;
        var decoder = this.dec;
        var reader = decoder.reader;
        var node = new BinaryRga_1.BinaryRga(id);
        node.ingest(length, function () {
            var chunkId = _this.ts();
            var _a = tslib_1.__read(reader.b1vu56(), 2), deleted = _a[0], length = _a[1];
            if (deleted)
                return new BinaryRga_1.BinaryChunk(chunkId, length, undefined);
            var data = reader.buf(length);
            return new BinaryRga_1.BinaryChunk(chunkId, length, data);
        });
        return node;
    };
    Decoder.prototype.cArr = function (id, length) {
        var _this = this;
        var decoder = this.dec;
        var reader = decoder.reader;
        var node = new ArrayRga_1.ArrayRga(this.doc, id);
        node.ingest(length, function () {
            var chunkId = _this.ts();
            var _a = tslib_1.__read(reader.b1vu28(), 2), deleted = _a[0], length = _a[1];
            if (deleted)
                return new ArrayRga_1.ArrayChunk(chunkId, length, undefined);
            var data = [];
            for (var i = 0; i < length; i++)
                data.push(_this.ts());
            return new ArrayRga_1.ArrayChunk(chunkId, length, data);
        });
        return node;
    };
    return Decoder;
}());
exports.Decoder = Decoder;
