"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.JsonPatch = void 0;
var ArrayRga_1 = require("../types/rga-array/ArrayRga");
var deepEqual_1 = require("../../json-equal/deepEqual");
var json_pointer_1 = require("../../json-pointer");
var ObjectLww_1 = require("../types/lww-object/ObjectLww");
var util_1 = require("../../json-pointer/util");
var clock_1 = require("../../json-crdt-patch/clock");
var JsonPatch = (function () {
    function JsonPatch(model) {
        this.model = model;
    }
    JsonPatch.prototype.apply = function (ops) {
        var length = ops.length;
        for (var i = 0; i < length; i++)
            this.applyOp(ops[i]);
        return this;
    };
    JsonPatch.prototype.applyOp = function (op) {
        switch (op.op) {
            case 'add':
                this.applyOpAdd(op);
                break;
            case 'remove':
                this.applyRemove(op);
                break;
            case 'replace':
                this.applyReplace(op);
                break;
            case 'move':
                this.applyMove(op);
                break;
            case 'copy':
                this.applyCopy(op);
                break;
            case 'test':
                this.applyTest(op);
                break;
            case 'str_ins':
                this.applyStrIns(op);
                break;
            case 'str_del':
                this.applyStrDel(op);
                break;
            default:
                throw new Error('UNKNOWN_OP');
        }
        this.model.api.apply();
        return this;
    };
    JsonPatch.prototype.builder = function () {
        return this.model.api.builder;
    };
    JsonPatch.prototype.applyOpAdd = function (op) {
        var builder = this.builder();
        var steps = (0, util_1.toPath)(op.path);
        if (!steps.length)
            this.setRoot(op.value);
        else {
            var objSteps = steps.slice(0, steps.length - 1);
            var node = this.model.api.find(objSteps);
            var key = steps[steps.length - 1];
            if (node instanceof ObjectLww_1.ObjectLww) {
                builder.setKeys(node.id, [[String(key), builder.json(op.value)]]);
            }
            else if (node instanceof ArrayRga_1.ArrayRga) {
                var value = builder.json(op.value);
                if (key === '-') {
                    var length_1 = node.length();
                    var after = node.find(length_1 - 1) || node.id;
                    builder.insArr(node.id, after, [value]);
                }
                else {
                    var index = ~~key;
                    if ('' + index !== key)
                        throw new Error('INVALID_INDEX');
                    if (!index)
                        builder.insArr(node.id, node.id, [value]);
                    else {
                        var after = node.find(index - 1);
                        if (!after)
                            throw new Error('NOT_FOUND');
                        builder.insArr(node.id, after, [value]);
                    }
                }
            }
            else
                throw new Error('NOT_FOUND');
        }
    };
    JsonPatch.prototype.applyRemove = function (op) {
        var builder = this.builder();
        var steps = (0, util_1.toPath)(op.path);
        if (!steps.length)
            this.setRoot(null);
        else {
            var objSteps = steps.slice(0, steps.length - 1);
            var node = this.model.api.find(objSteps);
            var key = steps[steps.length - 1];
            if (node instanceof ObjectLww_1.ObjectLww) {
                var stringKey = String(key);
                if (node.get(stringKey) === undefined)
                    throw new Error('NOT_FOUND');
                builder.setKeys(node.id, [[stringKey, builder.const(undefined)]]);
            }
            else if (node instanceof ArrayRga_1.ArrayRga) {
                var key_1 = steps[steps.length - 1];
                var index = ~~key_1;
                if ('' + index !== key_1)
                    throw new Error('INVALID_INDEX');
                var id = node.find(index);
                if (!id)
                    throw new Error('NOT_FOUND');
                builder.del(node.id, [(0, clock_1.interval)(id, 0, 1)]);
            }
            else
                throw new Error('NOT_FOUND');
        }
    };
    JsonPatch.prototype.applyReplace = function (op) {
        var path = op.path, value = op.value;
        this.applyRemove({ op: 'remove', path: path });
        this.applyOpAdd({ op: 'add', path: path, value: value });
    };
    JsonPatch.prototype.applyMove = function (op) {
        var path = (0, util_1.toPath)(op.path);
        var from = (0, util_1.toPath)(op.from);
        if ((0, json_pointer_1.isChild)(from, path))
            throw new Error('INVALID_CHILD');
        var json = this.json(from);
        this.applyRemove({ op: 'remove', path: from });
        this.applyOpAdd({ op: 'add', path: path, value: json });
    };
    JsonPatch.prototype.applyCopy = function (op) {
        var path = (0, util_1.toPath)(op.path);
        var from = (0, util_1.toPath)(op.from);
        var json = this.json(from);
        this.applyOpAdd({ op: 'add', path: path, value: json });
    };
    JsonPatch.prototype.applyTest = function (op) {
        var path = (0, util_1.toPath)(op.path);
        var json = this.json(path);
        if (!(0, deepEqual_1.deepEqual)(json, op.value))
            throw new Error('TEST');
    };
    JsonPatch.prototype.applyStrIns = function (op) {
        var path = (0, util_1.toPath)(op.path);
        var node = this.model.api.str(path).node;
        var length = node.length();
        var after = op.pos ? node.find(length < op.pos ? length - 1 : op.pos - 1) : node.id;
        if (!after)
            throw new Error('OUT_OF_BOUNDS');
        this.builder().insStr(node.id, after, op.str);
    };
    JsonPatch.prototype.applyStrDel = function (op) {
        var _a;
        var path = (0, util_1.toPath)(op.path);
        var node = this.model.api.str(path).node;
        var length = node.length();
        if (length <= op.pos)
            return;
        var deletionLength = Math.min((_a = op.len) !== null && _a !== void 0 ? _a : op.str.length, length - op.pos);
        var range = node.findInterval(op.pos, deletionLength);
        if (!range)
            throw new Error('OUT_OF_BOUNDS');
        this.builder().del(node.id, range);
    };
    JsonPatch.prototype.get = function (steps) {
        var _a;
        var model = this.model;
        if (!steps.length)
            return model.view();
        else {
            var objSteps = steps.slice(0, steps.length - 1);
            var node = model.api.find(objSteps);
            var key = steps[steps.length - 1];
            if (node instanceof ObjectLww_1.ObjectLww) {
                return (_a = node.get(String(key))) === null || _a === void 0 ? void 0 : _a.view();
            }
            else if (node instanceof ArrayRga_1.ArrayRga) {
                var index = ~~key;
                if ('' + index !== key)
                    throw new Error('INVALID_INDEX');
                var arrNode = node.getNode(index);
                if (!arrNode)
                    throw new Error('NOT_FOUND');
                return arrNode.view();
            }
        }
        return undefined;
    };
    JsonPatch.prototype.json = function (steps) {
        var json = this.get(steps);
        if (json === undefined)
            throw new Error('NOT_FOUND');
        return json;
    };
    JsonPatch.prototype.setRoot = function (json) {
        var builder = this.builder();
        builder.root(builder.json(json));
    };
    return JsonPatch;
}());
exports.JsonPatch = JsonPatch;
