"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.AbstractRga = void 0;
var clock_1 = require("../../../json-crdt-patch/clock");
var Tree_1 = require("../../../util/trees/Tree");
var isUint8Array_1 = require("../../../util/buffers/isUint8Array");
var util_1 = require("../../../util/trees/util");
var idsCompare = function (ts1, ts2) { return ts1.sid - ts2.sid || ts1.time - ts2.time; };
var AbstractRga = (function () {
    function AbstractRga(id) {
        this.id = id;
        this.root = undefined;
        this.ids = new Tree_1.Tree(idsCompare);
    }
    AbstractRga.prototype.ins = function (ref, id, content) {
        var chunk = this.createChunk(id, content);
        var rootId = this.id;
        var isRootInsert = rootId.time === ref.time && rootId.sid === ref.sid;
        if (isRootInsert) {
            var first = this.first();
            if (!first)
                this.setRoot(chunk);
            else if ((0, clock_1.compare)(first.id, id) < 0)
                this.insertBefore(chunk, first);
            else {
                if ((0, clock_1.containsId)(first.id, first.span, id))
                    return;
                this.insertAfterRef(chunk, ref, first);
            }
        }
        else {
            var at = this.findById(ref);
            if (!at)
                return;
            var offsetInInsertAtChunk = ref.time - at.id.time;
            var atSpan = at.span;
            var needsSplit = offsetInInsertAtChunk + 1 < atSpan;
            if (needsSplit && (0, clock_1.containsId)(at.id, atSpan, id))
                return;
            if (needsSplit && (0, clock_1.compare)(id, (0, clock_1.tick)(ref, 1)) > 0) {
                this.insertInside(chunk, at, offsetInInsertAtChunk + 1);
            }
            else
                this.insertAfterRef(chunk, ref, at);
            this.splay(chunk, 15);
        }
    };
    AbstractRga.prototype.insAt = function (position, id, content) {
        var after = position ? this.find(position - 1) : this.id;
        if (!after)
            return undefined;
        this.ins(after, id, content);
        return after;
    };
    AbstractRga.prototype.delete = function (spans) {
        var length = spans.length;
        for (var i = 0; i < length; i++)
            this.deleteSpan(spans[i]);
        this.onViewChange();
    };
    AbstractRga.prototype.deleteSpan = function (span) {
        var len = span.span;
        var t1 = span.time;
        var t2 = t1 + len - 1;
        var start = this.findById(span);
        if (!start)
            return;
        var chunk = start;
        var last = chunk;
        while (chunk) {
            last = chunk;
            var id = chunk.id;
            var chunkSpan = chunk.span;
            var c1 = id.time;
            var c2 = c1 + chunkSpan - 1;
            if (chunk.del) {
                if (c2 >= t2)
                    break;
                chunk = chunk.s;
                continue;
            }
            var deleteStartsFromLeft = t1 <= c1;
            var deleteStartsInTheMiddle = t1 <= c2;
            if (deleteStartsFromLeft) {
                var deleteFullyContainsChunk = t2 >= c2;
                if (deleteFullyContainsChunk) {
                    chunk.delete();
                    this.dLen(chunk, -chunk.span);
                    if (t2 <= c2)
                        break;
                }
                else {
                    var range = t2 - c1 + 1;
                    var newChunk = this.split(chunk, range);
                    chunk.delete();
                    this.updateLenOne(newChunk);
                    this.dLen(chunk, -chunk.span);
                    break;
                }
            }
            else if (deleteStartsInTheMiddle) {
                var deleteContainsRightSide = t2 >= c2;
                if (deleteContainsRightSide) {
                    var offset = t1 - c1;
                    var newChunk = this.split(chunk, offset);
                    newChunk.delete();
                    newChunk.len = newChunk.r ? newChunk.r.len : 0;
                    this.dLen(chunk, -newChunk.span);
                    if (t2 <= c2)
                        break;
                }
                else {
                    var right = this.split(chunk, t2 - c1 + 1);
                    var mid = this.split(chunk, t1 - c1);
                    mid.delete();
                    this.updateLenOne(right);
                    this.updateLenOne(mid);
                    this.dLen(chunk, -mid.span);
                    break;
                }
            }
            chunk = chunk.s;
        }
        if (last)
            this.mergeTombstones2(start, last);
    };
    AbstractRga.prototype.find = function (position) {
        var curr = this.root;
        while (curr) {
            var l = curr.l;
            var leftLength = l ? l.len : 0;
            var span = void 0;
            if (position < leftLength)
                curr = l;
            else if (curr.del) {
                position -= leftLength;
                curr = curr.r;
            }
            else if (position < leftLength + (span = curr.span)) {
                var ticks = position - leftLength;
                var id = curr.id;
                return !ticks ? id : new clock_1.Timestamp(id.sid, id.time + ticks);
            }
            else {
                position -= leftLength + span;
                curr = curr.r;
            }
        }
    };
    AbstractRga.prototype.findChunk = function (position) {
        var curr = this.root;
        while (curr) {
            var l = curr.l;
            var leftLength = l ? l.len : 0;
            var span = void 0;
            if (position < leftLength)
                curr = l;
            else if (curr.del) {
                position -= leftLength;
                curr = curr.r;
            }
            else if (position < leftLength + (span = curr.span)) {
                return [curr, position - leftLength];
            }
            else {
                position -= leftLength + span;
                curr = curr.r;
            }
        }
    };
    AbstractRga.prototype.findInterval = function (position, length) {
        var ranges = [];
        if (!length)
            return ranges;
        var curr = this.root;
        var offset = 0;
        while (curr) {
            var leftLength = curr.l ? curr.l.len : 0;
            if (position < leftLength)
                curr = curr.l;
            else if (curr.del) {
                position -= leftLength;
                curr = curr.r;
            }
            else if (position < leftLength + curr.span) {
                offset = position - leftLength;
                break;
            }
            else {
                position -= leftLength + curr.span;
                curr = curr.r;
            }
        }
        if (!curr)
            return ranges;
        if (curr.span >= length + offset) {
            var id_1 = curr.id;
            ranges.push((0, clock_1.tss)(id_1.sid, id_1.time + offset, length));
            return ranges;
        }
        var len = curr.span - offset;
        var id = curr.id;
        ranges.push((0, clock_1.tss)(id.sid, id.time + offset, len));
        length -= len;
        curr = this.next(curr);
        if (!curr)
            return ranges;
        do {
            if (curr.del)
                continue;
            var id_2 = curr.id;
            var span = curr.span;
            if (length <= span) {
                ranges.push((0, clock_1.tss)(id_2.sid, id_2.time, length));
                return ranges;
            }
            ranges.push((0, clock_1.tss)(id_2.sid, id_2.time, span));
            length -= span;
        } while ((curr = this.next(curr)) && length > 0);
        return ranges;
    };
    AbstractRga.prototype.first = function () {
        var curr = this.root;
        while (curr)
            if (curr.l)
                curr = curr.l;
            else
                return curr;
        return curr;
    };
    AbstractRga.prototype.next = function (curr) {
        if (curr.r) {
            curr = curr.r;
            while (curr.l)
                curr = curr.l;
            return curr;
        }
        var p = curr.p;
        while (p && p.r === curr) {
            curr = p;
            p = p.p;
        }
        return p;
    };
    AbstractRga.prototype.prev = function (curr) {
        if (curr.l) {
            curr = curr.l;
            while (curr.r)
                curr = curr.r;
            return curr;
        }
        var p = curr.p;
        while (p && p.l === curr) {
            curr = p;
            p = p.p;
        }
        return p;
    };
    AbstractRga.prototype.length = function () {
        var root = this.root;
        return root ? root.len : 0;
    };
    AbstractRga.prototype.size = function () {
        return this.ids.size;
    };
    AbstractRga.prototype.pos = function (chunk) {
        var p = chunk.p;
        if (!p)
            return chunk.l ? chunk.l.len : 0;
        return (this.pos(p) +
            (p.r === chunk ? (p.del ? 0 : p.span) : (chunk.del ? 0 : -chunk.span) + (chunk.r ? -chunk.r.len : 0)));
    };
    AbstractRga.prototype.setRoot = function (chunk) {
        this.root = chunk;
        this.insertId(chunk);
        this.onViewChange();
    };
    AbstractRga.prototype.insertBefore = function (chunk, before) {
        var l = before.l;
        before.l = chunk;
        chunk.l = l;
        chunk.p = before;
        if (l)
            l.p = chunk;
        this.updateLenOne(chunk);
        this.dLen(before, chunk.span);
        this.insertId(chunk);
        this.onViewChange();
    };
    AbstractRga.prototype.insertAfter = function (chunk, after) {
        var r = after.r;
        after.r = chunk;
        chunk.r = r;
        chunk.p = after;
        if (r)
            r.p = chunk;
        this.updateLenOne(chunk);
        this.dLen(after, chunk.span);
        this.insertId(chunk);
        this.onViewChange();
    };
    AbstractRga.prototype.insertAfterRef = function (chunk, ref, left) {
        var id = chunk.id;
        var sid = id.sid;
        var time = id.time;
        var isSplit = false;
        for (;;) {
            var leftId = left.id;
            var leftNextTick = leftId.time + left.span;
            if (!left.s) {
                isSplit = leftId.sid === sid && leftNextTick === time && leftNextTick - 1 === ref.time;
                if (isSplit)
                    left.s = chunk;
            }
            var right = this.next(left);
            if (!right)
                break;
            var cmp = (0, clock_1.compare)(right.id, id);
            if (cmp < 0)
                break;
            if (!cmp)
                return;
            left = right;
        }
        if (isSplit && !left.del) {
            this.mergeContent(left, chunk.data);
            left.s = undefined;
        }
        else
            this.insertAfter(chunk, left);
    };
    AbstractRga.prototype.mergeContent = function (chunk, content) {
        var span1 = chunk.span;
        chunk.merge(content);
        this.dLen(chunk, chunk.span - span1);
        this.onViewChange();
        return;
    };
    AbstractRga.prototype.insertInside = function (chunk, at, offset) {
        var p = at.p;
        var l = at.l;
        var r = at.r;
        var s = at.s;
        var len = at.len;
        var at2 = at.split(offset);
        at.s = at2;
        at2.s = s;
        at.l = at.r = at2.l = at2.r = undefined;
        at2.l = undefined;
        chunk.p = p;
        if (!l) {
            chunk.l = at;
            at.p = chunk;
        }
        else {
            chunk.l = l;
            l.p = chunk;
            var a = l.r;
            l.r = at;
            at.p = l;
            at.l = a;
            if (a)
                a.p = at;
        }
        if (!r) {
            chunk.r = at2;
            at2.p = chunk;
        }
        else {
            chunk.r = r;
            r.p = chunk;
            var b = r.l;
            r.l = at2;
            at2.p = r;
            at2.r = b;
            if (b)
                b.p = at2;
        }
        if (!p)
            this.root = chunk;
        else if (p.l === at)
            p.l = chunk;
        else
            p.r = chunk;
        this.updateLenOne(at);
        this.updateLenOne(at2);
        if (l)
            l.len = (l.l ? l.l.len : 0) + at.len + (l.del ? 0 : l.span);
        if (r)
            r.len = (r.r ? r.r.len : 0) + at2.len + (r.del ? 0 : r.span);
        chunk.len = len + chunk.span;
        var span = chunk.span;
        var curr = chunk.p;
        while (curr) {
            curr.len += span;
            curr = curr.p;
        }
        this.insertId(at2);
        this.insertIdFast(chunk);
        this.onViewChange();
    };
    AbstractRga.prototype.split = function (chunk, ticks) {
        var s = chunk.s;
        var newChunk = chunk.split(ticks);
        var r = chunk.r;
        chunk.s = newChunk;
        newChunk.r = r;
        newChunk.s = s;
        chunk.r = newChunk;
        newChunk.p = chunk;
        this.insertId(newChunk);
        if (r)
            r.p = newChunk;
        return newChunk;
    };
    AbstractRga.prototype.mergeTombstones = function (ch1, ch2) {
        if (!ch1.del || !ch2.del)
            return false;
        var id1 = ch1.id;
        var id2 = ch2.id;
        if (id1.sid !== id2.sid)
            return false;
        if (id1.time + ch1.span !== id2.time)
            return false;
        ch1.s = ch2.s;
        ch1.span += ch2.span;
        this.deleteChunk(ch2);
        return true;
    };
    AbstractRga.prototype.mergeTombstones2 = function (start, end) {
        var curr = start;
        while (curr) {
            var next = this.next(curr);
            if (!next)
                break;
            var merged = this.mergeTombstones(curr, next);
            if (!merged) {
                if (next === end) {
                    if (next) {
                        var n = this.next(next);
                        if (n)
                            this.mergeTombstones(next, n);
                    }
                    break;
                }
                curr = curr.s;
            }
        }
        var prev = this.prev(start);
        if (prev)
            this.mergeTombstones(prev, start);
    };
    AbstractRga.prototype.deleteChunk = function (chunk) {
        this.deleteId(chunk);
        var p = chunk.p;
        var l = chunk.l;
        var r = chunk.r;
        if (!l && !r) {
            if (!p)
                this.root = undefined;
            else {
                if (p.l === chunk)
                    p.l = undefined;
                else
                    p.r = undefined;
            }
        }
        else if (l && r) {
            var mostRightChildFromLeft = l;
            while (mostRightChildFromLeft.r)
                mostRightChildFromLeft = mostRightChildFromLeft.r;
            mostRightChildFromLeft.r = r;
            r.p = mostRightChildFromLeft;
            var rLen = r.len;
            var curr = void 0;
            curr = mostRightChildFromLeft;
            if (!p) {
                this.root = l;
                l.p = undefined;
            }
            else {
                if (p.l === chunk)
                    p.l = l;
                else
                    p.r = l;
                l.p = p;
            }
            while (curr && curr !== p) {
                curr.len += rLen;
                curr = curr.p;
            }
        }
        else {
            var child = (l || r);
            child.p = p;
            if (!p)
                this.root = child;
            else if (p.l === chunk)
                p.l = child;
            else
                p.r = child;
        }
    };
    AbstractRga.prototype.insertId = function (chunk) {
        this.ids.set(chunk.id, chunk);
    };
    AbstractRga.prototype.insertIdFast = function (chunk) {
        this.ids.setFast(chunk.id, chunk);
    };
    AbstractRga.prototype.deleteId = function (chunk) {
        this.ids.delete(chunk.id);
    };
    AbstractRga.prototype.findById = function (id) {
        var chunk = this.ids.getOrNextLower(id);
        if (!chunk)
            return undefined;
        var chunkId = chunk.id;
        if (chunkId.sid !== id.sid)
            return undefined;
        var offset = id.time - chunkId.time;
        return offset < chunk.span ? chunk : undefined;
    };
    AbstractRga.prototype.dLen = function (chunk, delta) {
        while (chunk) {
            chunk.len += delta;
            chunk = chunk.p;
        }
    };
    AbstractRga.prototype.updateLenOne = function (chunk) {
        var l = chunk.l;
        var r = chunk.r;
        chunk.len = (chunk.del ? 0 : chunk.span) + (l ? l.len : 0) + (r ? r.len : 0);
    };
    AbstractRga.prototype.splay = function (chunk, repeat) {
        var p = chunk.p;
        if (!p)
            return;
        var pp = p.p;
        var l2 = p.l === chunk;
        if (!pp) {
            if (l2)
                this.root = (0, util_1.rSplay)(chunk, p);
            else
                this.root = (0, util_1.lSplay)(chunk, p);
            this.updateLenOne(p);
            this.updateLenOne(chunk);
            return;
        }
        var l1 = pp.l === p;
        if (l1 && l2)
            this.root = (0, util_1.llSplay)(this.root, chunk, p, pp);
        else if (!l1 && !l2)
            this.root = (0, util_1.rrSplay)(this.root, chunk, p, pp);
        else if (l1 && !l2)
            this.root = (0, util_1.lrSplay)(this.root, chunk, p, pp);
        else
            this.root = (0, util_1.rlSplay)(this.root, chunk, p, pp);
        this.updateLenOne(pp);
        this.updateLenOne(p);
        this.updateLenOne(chunk);
        if (repeat > 1)
            this.splay(chunk, repeat - 1);
    };
    AbstractRga.prototype.toString = function (tab) {
        if (tab === void 0) { tab = ''; }
        var view = this.view();
        var value = '';
        if ((0, isUint8Array_1.isUint8Array)(view))
            value += " { ".concat(view.toString(), " }");
        else if (typeof view === 'string')
            value += "{ ".concat(view.length > 32 ? JSON.stringify(view.substring(0, 32)) + '...' : JSON.stringify(view), " }");
        return "".concat(this.constructor.name, " ").concat((0, clock_1.toDisplayString)(this.id), " ").concat(value).concat(this.root ? this.toStringChunk(this.root, tab, '') : '[empty]');
    };
    AbstractRga.prototype.toStringChunk = function (chunk, tab, side) {
        var str = "\n".concat(tab).concat(this.formatChunk(chunk, tab, side));
        if (chunk.l)
            str += this.toStringChunk(chunk.l, tab + '  ', 'l');
        if (chunk.r)
            str += this.toStringChunk(chunk.r, tab + '  ', 'r');
        return str;
    };
    AbstractRga.prototype.formatChunk = function (chunk, tab, side) {
        var str = "".concat(side === 'l' ? '<-' : side === 'r' ? '->' : '└─', " ").concat(chunk.constructor.name, " ").concat((0, clock_1.toDisplayString)(chunk.id), "!").concat(chunk.span, " len:").concat(chunk.len);
        if (chunk.del)
            str += " [".concat(chunk.span, "]");
        else {
            if ((0, isUint8Array_1.isUint8Array)(chunk.data))
                str += " { ".concat(chunk.data.toString(), " }");
            else if (typeof chunk.data === 'string')
                str += " { ".concat(chunk.data.length > 32 ? JSON.stringify(chunk.data.substring(0, 32)) + '...' : JSON.stringify(chunk.data), " }");
        }
        if (chunk.s)
            str += " ".concat(chunk.s ? "---> ".concat((0, clock_1.toDisplayString)(chunk.s.id)) : '');
        return str;
    };
    AbstractRga.prototype.iterator = function () {
        var _this = this;
        var curr = this.first();
        return function () {
            var res = curr;
            if (curr)
                curr = _this.next(curr);
            return res;
        };
    };
    AbstractRga.prototype.ingest = function (size, next) {
        if (size < 1)
            return;
        var splitLeftChunks = new Map();
        this.root = this._ingest(size, function () {
            var chunk = next();
            var id = chunk.id;
            var key = id.sid + '.' + id.time;
            var split = splitLeftChunks.get(key);
            if (split) {
                split.s = chunk;
                splitLeftChunks.delete(key);
            }
            var nextStampAfterSpan = (0, clock_1.tick)(id, chunk.span);
            splitLeftChunks.set(nextStampAfterSpan.sid + '.' + nextStampAfterSpan.time, chunk);
            return chunk;
        });
    };
    AbstractRga.prototype._ingest = function (size, next) {
        var leftSize = size >> 1;
        var rightSize = size - leftSize - 1;
        var c1 = leftSize > 0 ? this._ingest(leftSize, next) : undefined;
        var c2 = next();
        if (c1) {
            c2.l = c1;
            c1.p = c2;
        }
        var c3 = rightSize > 0 ? this._ingest(rightSize, next) : undefined;
        if (c3) {
            c2.r = c3;
            c3.p = c2;
        }
        this.updateLenOne(c2);
        this.insertId(c2);
        return c2;
    };
    return AbstractRga;
}());
exports.AbstractRga = AbstractRga;
