"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.MsgPackEncoderFast = void 0;
var Writer_1 = require("../../util/buffers/Writer");
var MsgPackEncoderFast = (function () {
    function MsgPackEncoderFast(writer) {
        if (writer === void 0) { writer = new Writer_1.Writer(); }
        this.writer = writer;
    }
    MsgPackEncoderFast.prototype.encode = function (json) {
        this.writer.reset();
        this.encodeAny(json);
        return this.writer.flush();
    };
    MsgPackEncoderFast.prototype.encodeAny = function (json) {
        switch (json) {
            case null:
                return this.writer.u8(0xc0);
            case false:
                return this.writer.u8(0xc2);
            case true:
                return this.writer.u8(0xc3);
        }
        if (json instanceof Array)
            return this.encodeArray(json);
        switch (typeof json) {
            case 'number':
                return this.encodeNumber(json);
            case 'string':
                return this.encodeString(json);
            case 'object':
                return this.encodeObject(json);
        }
    };
    MsgPackEncoderFast.prototype.encodeFloat64 = function (num) {
        this.writer.u8f64(0xcb, num);
    };
    MsgPackEncoderFast.prototype.encodeUnsignedInteger = function (num) {
        var writer = this.writer;
        this.writer.ensureCapacity(5);
        var uint8 = writer.uint8;
        if (num <= 127) {
            uint8[writer.x++] = num;
        }
        else if (num <= 0xffff) {
            uint8[writer.x++] = 0xcd;
            writer.view.setUint16(writer.x, num);
            writer.x += 2;
        }
        else {
            uint8[writer.x++] = 0xce;
            writer.view.setUint32(writer.x, num);
            writer.x += 4;
        }
    };
    MsgPackEncoderFast.prototype.encodeNegativeInteger = function (num) {
        var writer = this.writer;
        this.writer.ensureCapacity(5);
        var uint8 = writer.uint8;
        if (num >= -0x20) {
            uint8[writer.x++] = 0x100 + num;
        }
        else if (num >= -0x8000) {
            uint8[writer.x++] = 0xd1;
            writer.view.setInt16(writer.x, num);
            writer.x += 2;
        }
        else {
            uint8[writer.x++] = 0xd2;
            writer.view.setInt32(writer.x, num);
            writer.x += 4;
        }
    };
    MsgPackEncoderFast.prototype.encodeNumber = function (num) {
        if (num >>> 0 === num)
            return this.encodeUnsignedInteger(num);
        if (num >> 0 === num)
            return this.encodeNegativeInteger(num);
        this.encodeFloat64(num);
    };
    MsgPackEncoderFast.prototype.encodeNull = function () {
        this.writer.u8(0xc0);
    };
    MsgPackEncoderFast.prototype.encodeTrue = function () {
        this.writer.u8(0xc3);
    };
    MsgPackEncoderFast.prototype.encodeFalse = function () {
        this.writer.u8(0xc2);
    };
    MsgPackEncoderFast.prototype.encodeBoolean = function (bool) {
        if (bool)
            this.writer.u8(0xc3);
        else
            this.writer.u8(0xc2);
    };
    MsgPackEncoderFast.prototype.encodeStringHeader = function (length) {
        if (length <= 31)
            this.writer.u8(160 | length);
        else if (length <= 0xff) {
            this.writer.u8(0xd9);
            this.writer.u8(length);
        }
        else if (length <= 0xffff) {
            this.writer.u8(0xda);
            this.writer.u16(length);
        }
        else if (length <= 0xffffffff) {
            this.writer.u8(0xdb);
            this.writer.u32(length);
        }
    };
    MsgPackEncoderFast.prototype.encodeString = function (str) {
        var writer = this.writer;
        var length = str.length;
        var maxSize = length * 4;
        writer.ensureCapacity(5 + maxSize);
        var uint8 = writer.uint8;
        var lengthOffset = writer.x;
        if (maxSize <= 31)
            writer.x++;
        else if (maxSize <= 0xff) {
            uint8[writer.x++] = 0xd9;
            lengthOffset = writer.x;
            writer.x++;
        }
        else if (maxSize <= 0xffff) {
            uint8[writer.x++] = 0xda;
            lengthOffset = writer.x;
            writer.x += 2;
        }
        else {
            uint8[writer.x++] = 0xdb;
            lengthOffset = writer.x;
            writer.x += 4;
        }
        var bytesWritten = this.writer.utf8(str);
        if (maxSize <= 31)
            uint8[lengthOffset] = 160 | bytesWritten;
        else if (maxSize <= 0xff)
            uint8[lengthOffset] = bytesWritten;
        else if (maxSize <= 0xffff)
            writer.view.setUint16(lengthOffset, bytesWritten);
        else
            writer.view.setUint32(lengthOffset, bytesWritten);
    };
    MsgPackEncoderFast.prototype.encodeAsciiString = function (str) {
        var length = str.length;
        this.encodeStringHeader(length);
        this.writer.ensureCapacity(length);
        var uint8 = this.writer.uint8;
        var offset = this.writer.x;
        var pos = 0;
        while (pos < length)
            uint8[offset++] = str.charCodeAt(pos++);
        this.writer.x = offset;
    };
    MsgPackEncoderFast.prototype.encodeArrayHeader = function (length) {
        if (length <= 15)
            this.writer.u8(144 | length);
        else if (length <= 0xffff) {
            this.writer.u8u16(0xdc, length);
        }
        else if (length <= 0xffffffff) {
            this.writer.u8u32(0xdd, length);
        }
    };
    MsgPackEncoderFast.prototype.encodeArray = function (arr) {
        var length = arr.length;
        if (length <= 15)
            this.writer.u8(144 | length);
        else if (length <= 0xffff) {
            this.writer.u8u16(0xdc, length);
        }
        else if (length <= 0xffffffff) {
            this.writer.u8u32(0xdd, length);
        }
        else
            return;
        for (var i = 0; i < length; i++)
            this.encodeAny(arr[i]);
    };
    MsgPackEncoderFast.prototype.encodeObjectHeader = function (length) {
        if (length <= 15)
            this.writer.u8(128 | length);
        else if (length <= 0xffff) {
            this.writer.u8u16(0xde, length);
        }
        else if (length <= 0xffffffff) {
            this.writer.u8u32(0xdf, length);
        }
    };
    MsgPackEncoderFast.prototype.encodeObject = function (obj) {
        var keys = Object.keys(obj);
        var length = keys.length;
        this.encodeObjectHeader(length);
        for (var i = 0; i < length; i++) {
            var key = keys[i];
            this.encodeString(key);
            this.encodeAny(obj[key]);
        }
    };
    MsgPackEncoderFast.prototype.encodeExtHeader = function (type, length) {
        switch (length) {
            case 1:
                this.writer.u16((0xd4 << 8) | type);
                break;
            case 2:
                this.writer.u16((0xd5 << 8) | type);
                break;
            case 4:
                this.writer.u16((0xd6 << 8) | type);
                break;
            case 8:
                this.writer.u16((0xd7 << 8) | type);
                break;
            case 16:
                this.writer.u16((0xd8 << 8) | type);
                break;
            default:
                if (length <= 0xff) {
                    this.writer.u16((0xc7 << 8) | length);
                    this.writer.u8(type);
                }
                else if (length <= 0xffff) {
                    this.writer.u8u16(0xc8, length);
                    this.writer.u8(type);
                }
                else if (length <= 0xffffffff) {
                    this.writer.u8u32(0xc9, length);
                    this.writer.u8(type);
                }
        }
    };
    MsgPackEncoderFast.prototype.encodeExt = function (ext) {
        var type = ext.tag, buf = ext.val;
        var length = buf.byteLength;
        this.encodeExtHeader(type, length);
        this.writer.buf(buf, length);
    };
    MsgPackEncoderFast.prototype.encodeBinaryHeader = function (length) {
        if (length <= 0xff)
            this.writer.u16((0xc4 << 8) | length);
        else if (length <= 0xffff) {
            this.writer.u8u16(0xc5, length);
        }
        else if (length <= 0xffffffff) {
            this.writer.u8u32(0xc6, length);
        }
    };
    MsgPackEncoderFast.prototype.encodeBinary = function (buf) {
        var length = buf.byteLength;
        this.encodeBinaryHeader(length);
        this.writer.buf(buf, length);
    };
    return MsgPackEncoderFast;
}());
exports.MsgPackEncoderFast = MsgPackEncoderFast;
