"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Writer = void 0;
var EMPTY_UINT8 = new Uint8Array([]);
var EMPTY_VIEW = new DataView(EMPTY_UINT8.buffer);
var hasBuffer = typeof Buffer !== undefined;
var utf8Write = hasBuffer
    ? Buffer.prototype.utf8Write
    : null;
var from = hasBuffer ? Buffer.from : null;
var textEncoder = typeof TextEncoder !== 'undefined' ? new TextEncoder() : null;
var Writer = (function () {
    function Writer(allocSize) {
        if (allocSize === void 0) { allocSize = 65536; }
        this.allocSize = allocSize;
        this.view = EMPTY_VIEW;
        this.x = 0;
        this.uint8 = new Uint8Array(allocSize);
    }
    Writer.prototype.grow = function (size) {
        var newUint8 = new Uint8Array(size);
        newUint8.set(this.uint8);
        this.uint8 = newUint8;
        this.view = new DataView(newUint8.buffer);
    };
    Writer.prototype.ensureCapacity = function (capacity) {
        var size = this.x + capacity;
        var len = this.uint8.byteLength;
        if (len < size)
            this.grow(Math.max(size, len * 4));
    };
    Writer.prototype.reset = function () {
        var uint8 = this.uint8;
        this.view = new DataView(uint8.buffer, uint8.byteOffset, uint8.byteLength);
        this.x = 0;
    };
    Writer.prototype.flush = function () {
        var uint8 = this.uint8;
        var resultLength = this.x;
        var result = uint8.subarray(0, resultLength);
        var totalLength = uint8.byteLength;
        var bytesLeft = totalLength - resultLength;
        this.uint8 = bytesLeft > resultLength ? uint8.subarray(resultLength) : new Uint8Array(this.allocSize);
        return result;
    };
    Writer.prototype.u8 = function (char) {
        this.ensureCapacity(1);
        this.uint8[this.x++] = char;
    };
    Writer.prototype.u16 = function (word) {
        this.ensureCapacity(2);
        this.view.setUint16(this.x, word);
        this.x += 2;
    };
    Writer.prototype.u32 = function (dword) {
        this.ensureCapacity(4);
        this.view.setUint32(this.x, dword);
        this.x += 4;
    };
    Writer.prototype.i32 = function (dword) {
        this.ensureCapacity(4);
        this.view.setInt32(this.x, dword);
        this.x += 4;
    };
    Writer.prototype.u64 = function (qword) {
        this.ensureCapacity(8);
        this.view.setBigUint64(this.x, BigInt(qword));
        this.x += 8;
    };
    Writer.prototype.f64 = function (float) {
        this.ensureCapacity(8);
        this.view.setFloat64(this.x, float);
        this.x += 8;
    };
    Writer.prototype.u8u16 = function (u8, u16) {
        this.ensureCapacity(3);
        this.uint8[this.x++] = u8;
        this.uint8[this.x++] = u16 >>> 8;
        this.uint8[this.x++] = u16 & 0xff;
    };
    Writer.prototype.u8u32 = function (u8, u32) {
        this.ensureCapacity(5);
        this.uint8[this.x++] = u8;
        this.view.setUint32(this.x, u32);
        this.x += 4;
    };
    Writer.prototype.u8u64 = function (u8, u64) {
        this.ensureCapacity(9);
        this.uint8[this.x++] = u8;
        this.view.setBigUint64(this.x, BigInt(u64));
        this.x += 8;
    };
    Writer.prototype.u8f32 = function (u8, f32) {
        this.ensureCapacity(5);
        this.uint8[this.x++] = u8;
        this.view.setFloat32(this.x, f32);
        this.x += 4;
    };
    Writer.prototype.u8f64 = function (u8, f64) {
        this.ensureCapacity(9);
        this.uint8[this.x++] = u8;
        this.view.setFloat64(this.x, f64);
        this.x += 8;
    };
    Writer.prototype.buf = function (buf, length) {
        this.ensureCapacity(length);
        this.uint8.set(buf, this.x);
        this.x += length;
    };
    Writer.prototype.utf8 = function (str) {
        var maxLength = str.length * 4;
        if (maxLength < 168)
            return this.utf8Native(str);
        if (utf8Write) {
            var writeLength = utf8Write.call(this.uint8, str, this.x, maxLength);
            this.x += writeLength;
            return writeLength;
        }
        else if (from) {
            var uint8 = this.uint8;
            var offset = uint8.byteOffset + this.x;
            var buf = from(uint8.buffer).subarray(offset, offset + maxLength);
            var writeLength = buf.write(str, 0, maxLength, 'utf8');
            this.x += writeLength;
            return writeLength;
        }
        else if (maxLength > 1024 && textEncoder) {
            var writeLength = textEncoder.encodeInto(str, this.uint8.subarray(this.x, this.x + maxLength)).written;
            this.x += writeLength;
            return writeLength;
        }
        return this.utf8Native(str);
    };
    Writer.prototype.utf8Native = function (str) {
        var length = str.length;
        var uint8 = this.uint8;
        var offset = this.x;
        var pos = 0;
        while (pos < length) {
            var value = str.charCodeAt(pos++);
            if ((value & 0xffffff80) === 0) {
                uint8[offset++] = value;
                continue;
            }
            else if ((value & 0xfffff800) === 0) {
                uint8[offset++] = ((value >> 6) & 0x1f) | 0xc0;
            }
            else {
                if (value >= 0xd800 && value <= 0xdbff) {
                    if (pos < length) {
                        var extra = str.charCodeAt(pos);
                        if ((extra & 0xfc00) === 0xdc00) {
                            pos++;
                            value = ((value & 0x3ff) << 10) + (extra & 0x3ff) + 0x10000;
                        }
                    }
                }
                if ((value & 0xffff0000) === 0) {
                    uint8[offset++] = ((value >> 12) & 0x0f) | 0xe0;
                    uint8[offset++] = ((value >> 6) & 0x3f) | 0x80;
                }
                else {
                    uint8[offset++] = ((value >> 18) & 0x07) | 0xf0;
                    uint8[offset++] = ((value >> 12) & 0x3f) | 0x80;
                    uint8[offset++] = ((value >> 6) & 0x3f) | 0x80;
                }
            }
            uint8[offset++] = (value & 0x3f) | 0x80;
        }
        var writeLength = offset - this.x;
        this.x = offset;
        return writeLength;
    };
    return Writer;
}());
exports.Writer = Writer;
